/* codegen.c */

#include "opldefs.h"
#include <stdio.h>
#include <stdlib.h>
                  
extern int verbose;
extern int generate_debug;
extern int compile_only;
extern int translate_only;
extern int throwback;

FILE *out;

void putindent( int indent );
void putline( const char* line );
void puttext( const char* line );
void outputCProgram        ( program *prog );
void outputCProcedureProto ( procedure *proc );
void outputCProcedureBody  ( procedure *proc );
void outputCStatementList  ( statementList *pStmts, int indent );
void outputCStatement      ( statement *stmt, int indent );
void outputCExpression     ( expression *expr );
void outputCIdentType      ( ident* id );


void outputCProgram( program *prog )
{
  procedureList *pProc;

  if( !prog )
    return;

  putline( "#include \"OPLlib.h\"" );
  putline( "#include <string.h>" );
  putline( "" );

  for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
  {
    if( pProc->head )
      outputCProcedureProto( pProc->head );
  }

  putline( "" );

  for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
  {
    if( pProc->head )
    {
      outputCProcedureBody( pProc->head );
      putline( "" );
    }
  }

  putline( "int main( int argc, char **argv )" );
  putline( "{" );
  putline( "  if( OPLlib_init( argv[0] ) )" );
  putline( "  {" );
  if( prog->pProcs && prog->pProcs->head )
  {
    fprintf( out, "    %s();\n", prog->pProcs->head->name->name );
  }
  putline( "    OPLlib_run();" );
  putline( "  }" );
  putline( "  return 0;" );
  putline( "}" );
}



void outputCProcedureProto( procedure *proc )
{
  statementList *pStmt;
  identList *pVars;
  identList *pArg;
  expressionList *pArray;
  int first = TRUE;

  if( !proc )
    return;

  outputCIdentType( proc->name );
  fprintf( out, " %s( ", proc->name->name );

  if( proc->pArgs )
  {
    for( pArg = proc->pArgs; pArg; pArg = pArg->tail )
    {
      if( pArg->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputCIdentType( pArg->head );
        fprintf( out, " %s", pArg->head->name );
      }
      first = FALSE;
    }
  }
  else
  {
    fprintf( out, "void" );
  }

  fprintf( out, " );\n" );

  for( pStmt = proc->pStmts; pStmt; pStmt = pStmt->tail )
  {
    if( pStmt->head )
    {
      switch( pStmt->head->kind )
      {
      case global_stmt:
        for( pVars = pStmt->head->data.global_and_local_part.pVars;
             pVars; pVars = pVars->tail )
        {
          if( pVars->head )
          {
            outputCIdentType( pVars->head );
            fprintf( out, " %s", pVars->head->name );
            if( pVars->head->pArrays )
            {
              for( pArray = pVars->head->pArrays;
                   pArray; pArray = pArray->tail )
              {
                if( pArray->head && ( pVars->head->type != string_id ) )
                {
                  fprintf( out, "[ " );
                  outputCExpression( pArray->head );
                  fprintf( out, " ]" );
                }
              }
            }
            fprintf( out, ";\n" );
          }
        }
        break;
      }
    }
  }
}



void outputCProcedureBody( procedure *proc )
{
  identList *pArg;
  int first = TRUE;

  if( !proc )
    return;

  outputCIdentType( proc->name );
  fprintf( out, " %s( ", proc->name->name );

  if( proc->pArgs )
  {
    for( pArg = proc->pArgs; pArg; pArg = pArg->tail )
    {
      if( pArg->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputCIdentType( pArg->head );
        fprintf( out, " %s", pArg->head->name );
      }
      first = FALSE;
    }
  }
  else
  {
    fprintf( out, "void" );
  }

  fprintf( out, " )\n" );
  putline( "{" );
  outputCStatementList( proc->pStmts, 1 );

  if( proc->name->type != untyped_id )
    putline( "  return 0;" );

  putline( "}" );
}
 


void outputCStatementList( statementList *pStmts, int indent )
{
  statementList *pStmt;

  if( !pStmts )
    return;

  for( pStmt = pStmts; pStmt; pStmt = pStmt->tail )
  {
    if( pStmt->head )
      outputCStatement( pStmt->head, indent );
  }
}



void outputCStatement( statement *stmt, int indent )
{
  expressionList *pExpr;
  identList *pVars;
  expressionList *pArray;
  int first = TRUE;

  if( !stmt )
    return;

  switch( stmt->kind )
  {
  case if_stmt:
    putindent( indent ); puttext( "if( " );
    outputCExpression( stmt->data.if_part.condition );
    putline( " )" );
    if( stmt->data.if_part.pThen )
    {
      if( stmt->data.if_part.pThen->tail
          || ( stmt->data.if_part.pThen->head
               && ( stmt->data.if_part.pThen->head->kind == if_stmt ) ) )
      {
        putindent( indent ); putline( "{" );
        outputCStatementList( stmt->data.if_part.pThen, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputCStatementList( stmt->data.if_part.pThen, indent+1 );
    }
    if( stmt->data.if_part.pElse )
    {
      putindent( indent ); putline( "else" );
      if( stmt->data.if_part.pElse->tail
          || ( stmt->data.if_part.pElse->head
               && ( stmt->data.if_part.pElse->head->kind == if_stmt ) ) )
      {
        putindent( indent ); putline( "{" );
        outputCStatementList( stmt->data.if_part.pElse, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputCStatementList( stmt->data.if_part.pElse, indent+1 );
    }
    break;
  case while_stmt:
    putindent( indent ); puttext( "while( " );
    outputCExpression( stmt->data.while_and_do_part.condition );
    putline( " )" );
    if( stmt->data.while_and_do_part.pBody )
    {
      if( stmt->data.while_and_do_part.pBody->tail )
      {
        putindent( indent ); putline( "{" );
        outputCStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputCStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
    }
    break;
  case do_stmt:
    putindent( indent ); putline( "do" );
    if( stmt->data.while_and_do_part.pBody )
    {
      if( stmt->data.while_and_do_part.pBody->tail )
      {
        putindent( indent ); putline( "{" );
        outputCStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputCStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
    }
    putindent( indent ); puttext( "while( " );
    outputCExpression( stmt->data.while_and_do_part.condition );
    putline( " );" );
    break;
  case assignment_stmt:
    putindent( indent );
    if( ( stmt->data.assignment_part.name->type == string_id )
        && !stmt->data.assignment_part.name->pArrays )
    {
      fprintf( out, "%s", stmt->data.assignment_part.name->name );
      fprintf( out, " = strdup(" );
      outputCExpression( stmt->data.assignment_part.value );
      fprintf( out, " )" );
    }
    else
    {
      fprintf( out, "%s", stmt->data.assignment_part.name->name );
      if( stmt->data.assignment_part.name->pArrays )
      {
        for( pArray = stmt->data.assignment_part.name->pArrays;
             pArray; pArray = pArray->tail )
        {
          if( pArray->head )
          {
            fprintf( out, "[ " );
            outputCExpression( pArray->head );
            fprintf( out, " ]" );
          }
        }
      }
      fprintf( out, " = " );
      outputCExpression( stmt->data.assignment_part.value );
    }
    fprintf( out, ";\n" );
    break;
  case global_stmt:
    break;
  case local_stmt:
    for( pVars = stmt->data.global_and_local_part.pVars;
         pVars; pVars = pVars->tail )
    {
      if( pVars->head )
      {
        putindent( indent );
        outputCIdentType( pVars->head );
        fprintf( out, " %s", pVars->head->name );
        if( pVars->head->pArrays )
        {
          for( pArray = pVars->head->pArrays; pArray; pArray = pArray->tail )
          {
            if( pArray->head && ( pVars->head->type != string_id ) )
            {
              fprintf( out, "[ " );
              outputCExpression( pArray->head );
              fprintf( out, " ]" );
            }
          }
        }
        fprintf( out, ";\n" );
      }
    }
    break;
  case call_stmt:
    putindent( indent );
    if( stmt->data.call_part.opl_call )
      fprintf( out, "OPL_" );
    fprintf( out, "%s( ", stmt->data.call_part.name->name );
    for( pExpr = stmt->data.call_part.pArgs; pExpr; pExpr = pExpr->tail )
    {
      if( pExpr->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputCExpression( pExpr->head );
      }
      first = FALSE;
    }
    fprintf( out, " );\n" );
    break;
  case return_stmt:
    putindent( indent );
    fprintf( out, "return " );
    if( stmt->data.return_part.expr )
      outputCExpression( stmt->data.return_part.expr );
    fprintf( out, ";\n" );
   break;
  }
}
 


void outputCExpression( expression *expr )
{
  expressionList *pExpr;
  expressionList *pArray;
  int first = TRUE;

  if( !expr )
    return;

  switch( expr->kind )
  {
  case variable_expr:
    fprintf( out, "%s", expr->data.variable_part->name );
    if( expr->data.variable_part->pArrays )
    {
      for( pArray = expr->data.variable_part->pArrays;
           pArray; pArray = pArray->tail )
      {
        if( pArray->head )
        {
          fprintf( out, "[ " );
          outputCExpression( pArray->head );
          fprintf( out, " ]" );
        }
      }
    }
    break;
  case integer_expr:
    fprintf( out, "%d", expr->data.integer_part );
    break;
  case hexint_expr:
    fprintf( out, "0x%x", expr->data.integer_part );
    break;
  case character_expr:
    fprintf( out, "'%c'", expr->data.integer_part );
    break;
  case floating_point_expr:
    fprintf( out, "%f", expr->data.floating_point_part );
    break;
  case long_integer_expr:
    fprintf( out, "%d", (int) expr->data.long_integer_part );
    break;
  case string_expr:
    fprintf( out, "\"%s\"", expr->data.string_part );
    break;
  case addition_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " + " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case subtraction_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " - " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case multiplication_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " * " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case division_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " / " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case equality_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " == " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case inequality_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " != " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case lessthan_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " < " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case greaterthan_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " > " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case lessequal_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " <= " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case greaterequal_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " >= " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case and_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " && " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case or_expr:
    outputCExpression( expr->data.binary_op_part.left );
    fprintf( out, " || " );
    outputCExpression( expr->data.binary_op_part.right );
    break;
  case uminus_expr:
    fprintf( out, "- " );
    outputCExpression( expr->data.unary_op_part.expr );
    break;
  case parenthesis_expr:
    fprintf( out, "( " );
    outputCExpression( expr->data.unary_op_part.expr );
    fprintf( out, " )" );
    break;
  case call_expr:
    if( expr->data.call_part.opl_call )
      fprintf( out, "OPL_" );
    fprintf( out, "%s( ", expr->data.call_part.name->name );
    for( pExpr = expr->data.call_part.pArgs; pExpr; pExpr = pExpr->tail )
    {
      if( pExpr->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputCExpression( pExpr->head );
      }
      first = FALSE;
    }
    fprintf( out, " )" );
    break;
  }
}



void outputCIdentType( ident* id )
{
  switch( id->type )
  {
  case untyped_id:
    fprintf( out, "void" );
    break;
  case integer_id:
    fprintf( out, "int" );
    break;
  case long_integer_id:
    fprintf( out, "long" );
    break;
  case floating_point_id:
    fprintf( out, "double" );
    break;
  case string_id:
    /*if( id->pArrays )
      fprintf( out, "char" );
    else*/
      fprintf( out, "char*" );
    break;
  }
}



void putline( const char* line )
{
  if( !line )
    return;

  fprintf( out, "%s\n", line );
}


void puttext( const char* line )
{
  if( !line )
    return;

  fprintf( out, line );
}


void putindent( int indent )
{
  int n;

  for( n = 0; n < indent; n++ )
    fprintf( out, "  " );    
}



void generateCCode( program *prog )
{
  char command[1024];

  if( !prog )
    return;

  out = fopen( "program.c", "w" );

  if( !out )
    return;

  outputCProgram( prog );

  fclose( out );

  if( !translate_only )
  {
    sprintf( command, "cc -I<OPL2C$Dir> %s%s%s-o m.program m.program.c%s\n",
             (compile_only ? "-c " : "" ), (generate_debug ? "-g " : "" ),
             (throwback    ? "-throwback " : "" ),
             (compile_only ? "" : " <OPL2C$Dir>.o.OPLlib C:RISC_OSLib.o" ) );

    if( verbose )
      fprintf( stderr, command );

    system( command );
  }
}

