/* scanner for a OPL */

%{
#include "opldefs.h"
#include <string.h>
#include <stdlib.h>

#ifndef __GNUC__
extern char *strdup( const char *s );
extern char *strndup( const char *s, int n );
#else
extern char *strndup( const char *s, int n );
#endif

#include "opldefs.h"
#include "oplpars.h"

#ifdef __riscos
#include "throwback.h"
#endif


#define RETURN_OPL_FUNCTION( x, t )         { yylval.id = new_ident( x, t, NULL ); return OPL_FUNCTION; }
#define RETURN_OPL_PROCEDURE( x )           { yylval.string = x; return OPL_PROCEDURE;  }
#define RETURN_OPL_PROCEDURE_NO_ARGS( x )   { yylval.string = x; return OPL_PROCEDURE_NO_ARGS;  }
#define RETURN_OPL_PRINT_PROC( x )          { yylval.string = x; return OPL_PRINT_PROC; }

extern void opl2c_throwback_error( int line, int level, const char* error );

/* global line count */
int _lineNo = 1;
%}

identifier      [A-Za-z][A-Za-z0-9]*
      
%x cmnt
%x strcon

%%
                char string_buf[255];
                char *string_buf_ptr;

[ \t]+          /* eat up whitespace */

\n              ++_lineNo;

REM[^\n]*       /* eat up comments */

\"              {
                  string_buf_ptr = string_buf;
                  BEGIN(strcon);
                }

<strcon>\"      { /* saw closing quote - all done */
                  BEGIN(INITIAL);
                  *string_buf_ptr = '\0';
                  yylval.string = strdup( string_buf );
                  return STRING;
                }

<strcon>\n      yyerror( "unterminated string constant" );

<strcon>\\     {
                 *string_buf_ptr++ = '\\';
                 *string_buf_ptr++ = '\\';
               }

<strcon>[^\\\n\"] { /* + */
                  *string_buf_ptr++ = yytext[0];
                  /* char *yytext_ptr = yytext;
                  while ( *yytext_ptr )
                    *string_buf_ptr++ = *yytext_ptr++; */
                }

ASC             RETURN_OPL_FUNCTION( "asc", integer_id )
AT              RETURN_OPL_PROCEDURE( "at" )
APP             return APP;
AND             return AND;
BUSY            RETURN_OPL_PROCEDURE( "busy" )
"CHR$"          RETURN_OPL_FUNCTION( "chr", string_id )
CURSOR          RETURN_OPL_PROCEDURE( "cursor" )
DEFAULTWIN      RETURN_OPL_PROCEDURE( "defaultWin" )
dBUTTONS        RETURN_OPL_PROCEDURE( "dButtons" )
dCHOICE         RETURN_OPL_PROCEDURE( "dChoice" )
dEDIT           RETURN_OPL_PROCEDURE( "dEdit" )
dFLOAT          RETURN_OPL_PROCEDURE( "dFloat" )
DIALOG          RETURN_OPL_FUNCTION( "dialog", integer_id )
dINIT           RETURN_OPL_PROCEDURE( "dInit" ) /* can have no args */
DO              return DO;
dTEXT           RETURN_OPL_PROCEDURE( "dText" )
ELSE            return ELSE;
ELSEIF          return ELSEIF;
ENDA            return ENDA;
ENDP            return ENDP;
ENDIF           return ENDIF;
ENDWH           return ENDWH;
EXIT            RETURN_OPL_PROCEDURE_NO_ARGS( "exit" )
FONT            RETURN_OPL_PROCEDURE( "font" )
gAT             RETURN_OPL_PROCEDURE( "gAt" )
gBOX            RETURN_OPL_PROCEDURE( "gBox" )
gBORDER         RETURN_OPL_PROCEDURE( "gBorder" )
gCLOSE          RETURN_OPL_PROCEDURE( "gClose" )
gCLS            RETURN_OPL_PROCEDURE_NO_ARGS( "gCls" )
gCOPY           RETURN_OPL_PROCEDURE( "gCopy" )
gCREATE         RETURN_OPL_FUNCTION( "gCreate", integer_id )
GET             RETURN_OPL_FUNCTION( "get", integer_id )
"GETCMD$"       RETURN_OPL_FUNCTION( "getcmd", string_id )
GETEVENT        RETURN_OPL_PROCEDURE( "getEvent" )
GIPRINT         RETURN_OPL_PROCEDURE( "giprint" )
gLINEBY         RETURN_OPL_PROCEDURE( "gLineBy" )
gLINETO         RETURN_OPL_PROCEDURE( "gLineTo" )
gLOADBIT        RETURN_OPL_FUNCTION( "gLoadBit", integer_id )
GLOBAL          return GLOBAL;
gFILL           RETURN_OPL_PROCEDURE( "gFill" )
gFONT           RETURN_OPL_PROCEDURE( "gFont" )
gGREY           RETURN_OPL_PROCEDURE( "gGrey" )
gMOVE           RETURN_OPL_PROCEDURE( "gMove" )
gPATT           RETURN_OPL_PROCEDURE( "gPatt" )
gPRINT          RETURN_OPL_PRINT_PROC( "gPrint" )
gSTYLE          RETURN_OPL_PROCEDURE( "gStyle" )
gSETWIN         RETURN_OPL_PROCEDURE( "gSetwin" )
gTMODE          RETURN_OPL_PROCEDURE( "gTMode" )
gUSE            RETURN_OPL_PROCEDURE( "gUse" )
IF              return IF;
"LEFT$"         RETURN_OPL_FUNCTION( "left", string_id )
LOCAL           return LOCAL;
LOCK            RETURN_OPL_PROCEDURE( "lock" )
mCARD           RETURN_OPL_PROCEDURE( "mCard" )
MENU            RETURN_OPL_FUNCTION( "menu", integer_id )
mINIT           RETURN_OPL_PROCEDURE_NO_ARGS( "mInit" )
"NUM$"          RETURN_OPL_FUNCTION( "num", string_id )
OFF             { yylval.integer = 0; return INTEGER; }
ON              { yylval.integer = 1; return INTEGER; }
OR              return OR;
PAUSE           RETURN_OPL_PROCEDURE( "pause" )
PRINT           RETURN_OPL_PRINT_PROC( "print"  )
PROC            return PROC;
RANDOMIZE       RETURN_OPL_PROCEDURE( "randomize" )
RETURN          return RETURN;
RND             RETURN_OPL_FUNCTION( "rnd", floating_point_id )
SECOND          RETURN_OPL_FUNCTION( "second", integer_id );
SQR             RETURN_OPL_FUNCTION( "sqr", floating_point_id );
STATUSWIN       RETURN_OPL_PROCEDURE( "statusWin" )
UNTIL           return UNTIL;
WHILE           return WHILE;

"<="            return LE;
">="            return GE;
"<>"            return NE;
"**"            return POWER;

{identifier}"%:" {
                  char* func = strndup( yytext, strlen(yytext)-1 );
                  func[ strlen(yytext)-2 ] = '_';
                  yylval.id = new_ident( func, integer_id, NULL );
                  return SUBROUTINE;
                }

{identifier}"%" {
                  yylval.id = new_ident( strndup( yytext, strlen(yytext)-1 ),
                                         integer_id, NULL );
                  return IDENTIFIER;
                }

{identifier}"$:" {
                  char* func = strndup( yytext, strlen(yytext)-1 );
                  func[ strlen(yytext)-2 ] = '_';
                  yylval.id = new_ident( func, string_id, NULL );
                  return SUBROUTINE;
                }

"new$"          {
                  yylval.id = new_ident( "new_", string_id, NULL );
                  return IDENTIFIER;
                }

{identifier}"$" {
                  yylval.id = new_ident( strndup( yytext, strlen(yytext)-1 ),
                                         string_id, NULL );
                  return IDENTIFIER;
                }

{identifier}"&:" {
                  char* func = strndup( yytext, strlen(yytext)-1 );
                  func[ strlen(yytext)-2 ] = '_';
                  yylval.id = new_ident( func, long_integer_id, NULL );
                  return SUBROUTINE;
                }

{identifier}"&" {
                  yylval.id = new_ident( strndup( yytext, strlen(yytext)-1 ),
                                         long_integer_id, NULL );
                  return IDENTIFIER;
                }

{identifier}":" {
                  char* func = strdup( yytext );
                  func[ strlen(yytext)-1 ] = '_';
                  yylval.id = new_ident( func, floating_point_id, NULL );
                  return SUBROUTINE;
                }

{identifier}    {
                  yylval.id = new_ident( strdup( yytext ),
                                         floating_point_id, NULL );
                  return IDENTIFIER;
                }

"$"[0-9a-f]+    {
                  yylval.integer = (int) strtol( yytext+1, NULL, 16 );
                  return HEXINT;
                }

"%".            {
                  yylval.integer = yytext[1];
                  return CHARACTER;
                }

[0-9]+"."[0-9]+ {
                  yylval.integer = atoi( yytext );
                  return INTEGER; /* FLOAT */
                }

[0-9]+          {
                  yylval.integer = atoi( yytext );
                  return INTEGER;
                }

.               return (int) yytext[0];

%%

extern const char *infile;

void yyerror (s)  /* Called by yyparse on error */
     char *s;
{
  char buf[255];

  fprintf( stderr, "\"%s\", line %d: %s with '%s'\n", infile, _lineNo, s, yytext);

  sprintf( buf, "%s with '%s'", s, yytext );
  opl2c_throwback_error( _lineNo, 1, buf );
}
